# Copyright (c) 2011 Cloudera, Inc. All rights reserved.
import time
from datetime import datetime
import os
import simplejson as json
import re
import psutil
from psutil._common import ntuple_io
import gzip
import zipfile
import tempfile
import util

from clusterstats.log.streaming.file import LogFile
from clusterstats.log.streaming.event_streamer import LogEventStreamer
from clusterstats.log.streaming.event_reader import LogFileEventReader
from clusterstats.log.streaming.event_parser import Log4JEventParser, HueEventParser, GLogEventParser

from clusterstats.log.streaming.file_filter import DirectoryFilter
from clusterstats.log.streaming.file_filter import FilePatternFilter
from clusterstats.log.streaming.file_filter import TimeRangeFilter
from clusterstats.log.streaming.event_filter import LogLevelFilter, HueLogLevelFilter, GLogLogLevelFilter
import clusterstats.common.chain as chain

from clusterstats.log.streaming.grep_stream_strategy import GrepStreamStrategy
from clusterstats.log.streaming.grep_stream_strategy import TimeoutException

TEN_SECONDS_IN_MILLIS=10000

VALID_LOG_LEVELS = set(["TRACE", "DEBUG", "INFO", "WARN", "ERROR", "FATAL"])

HUE = "HUE"
LOG4J = "LOG4J"
GLOG = "GLOG"

VALID_LOG_TYPES = set([LOG4J, HUE, GLOG])

LOG4J_HUE_LEVEL_MAPPING = {
  "TRACE": "DEBUG",
  "DEBUG": "DEBUG",
  "INFO": "INFO",
  "WARN": "WARNING",
  "ERROR": "ERROR",
  "FATAL": "CRITICAL"}

EVENT_PARSER_CLASSES = {
  LOG4J: Log4JEventParser,
  HUE: HueEventParser,
  GLOG: GLogEventParser
}

LOG_LEVEL_FILTER_CLASSES = {
  LOG4J: LogLevelFilter,
  HUE: HueLogLevelFilter,
  GLOG: GLogLogLevelFilter
}

# This defines the level at which GrepStreamStrategy will be used for filtering
# log levels
GREPSTREAM_LEVEL_THRESHOLD = "WARN"

GREPSTREAM_LEVEL_THRESHOLDS = {
  LOG4J: GREPSTREAM_LEVEL_THRESHOLD,
  HUE: LOG4J_HUE_LEVEL_MAPPING[GREPSTREAM_LEVEL_THRESHOLD],
  GLOG: GREPSTREAM_LEVEL_THRESHOLD
}

TIMEOUT_MESSAGE = "Timed out while search, results might not be complete"

JAVA_QUOTE_REGEX = r"\\Q(.*)\\E"
quote_regex = re.compile(JAVA_QUOTE_REGEX)

PREVIOUS = "previous"
NEXT = "next"

proc = psutil.Process(os.getpid())

def datetime_from_millis(s):
  """
  Parses a timestamp in milliseconds since epoch
  """
  if s is None:
    raise Exception("ERROR: timestamp must be specified.")

  try:
    dt = datetime.fromtimestamp(s/1000.0)
  except ValueError, e:
    raise Exception("ERROR: dates must be in milliseconds since epoch" + e.args[0])

  return dt

def millis_from_datetime(date):
  return int(date.strftime("%s")) * 1000 + date.microsecond / 1000


def parse_log_level(s):
  """
  "Parses" the log level, making sure that it is a valid one.
  """
  log_level = s.strip()
  if log_level not in VALID_LOG_LEVELS:
    raise Exception("ERROR: Log level must be one of the following: "
                    + ",".join(VALID_LOG_LEVELS) + ".")
  return log_level


def validate_log_dir(log_dir):
  if not os.path.isdir(log_dir):
    raise Exception("ERROR: log directory %s does not exist.  (Note: log \
directories for some roles are only created at start-up)" % (log_dir))
  return log_dir

def validate_log_type(log_type):
  if log_type not in VALID_LOG_TYPES:
    raise Exception("ERROR: log type %s is not supported" % (log_type))
  return log_type

def datetime_converter(obj):
  if isinstance(obj, datetime):
    return millis_from_datetime(obj)
  else:
    return obj

def get_stream_strategy(log_level, log_type, query, timeout):
  """
  Returns the stream strategy a log stream should use. If the given query is not
  the empty or None, the the strategy will be to only parse lines that contain
  the query.

  Without a query, our strategy is to parse lines containing the given log_level
  as long as the log_level is >= WARN. This improves log streaming performance,
  as most lines will not be parsed.
  """

  if query is not None:
    query = query.strip()

  if query:
    query_regex = get_query_regex(query)
    return GrepStreamStrategy(query_regex, timeout)
  else:
    log_level_filter_class = LOG_LEVEL_FILTER_CLASSES[log_type]
    level_index = log_level_filter_class.levels.index(log_level)
    log_level_threshold = GREPSTREAM_LEVEL_THRESHOLDS[log_type]
    if (level_index < log_level_filter_class.levels.index(log_level_threshold)):
      # Not worth using grepstream strategy for levels below WARN since we'll
      # be parsing most events anyway
      return None

    # Make regex to match all log levels >= log_level.
    regex = re.compile(
      make_log_level_regex_string(level_index, log_level_filter_class))
    return GrepStreamStrategy(regex, timeout)

def make_log_level_regex_string(level_index, log_level_filter_class):
  return "|".join(log_level_filter_class.levels[level_index:])

def get_log_streamer(log_type, log_dir, log_name, log_level,
                     start_datetime, end_datetime, query, timeout):
  """
  Returns a log streamer constructed with the given properties.

  log_type: must be one of "GLOG", "HUE" or "LOG4J"
  log_dir: the log directory from which to stream logs
  log_name: name of the log file from which to stream logs
  start_datetime: a datetime object for the earliest timestamp for a log line
  end_datetime: a datetime object for the latest timestamp for a log line
  query: when set to a string, the log streamer will only return lines that contain
         it. may be set to None.
  timeout: the time (in seconds since the epoch) at which the log streamer
           should raise a timeout
  """
  if log_type == HUE:
    log_level = LOG4J_HUE_LEVEL_MAPPING[log_level]

  stream_strategy = get_stream_strategy(log_level, log_type, query, timeout)
  event_reader = LogFileEventReader(EVENT_PARSER_CLASSES[log_type]())
  log_level_filter = LOG_LEVEL_FILTER_CLASSES[log_type](log_level)
  time_range_filter = TimeRangeFilter(start_datetime, end_datetime, event_reader)
  file_pattern_filter = FilePatternFilter(re.escape(log_name) + '*')

  filters = chain.build_chain([DirectoryFilter(), file_pattern_filter, time_range_filter])

  return LogEventStreamer(log_dir, event_reader, stream_strategy, filters, log_level_filter)

def search_logs(start_time,
                end_time,
                log_types,
                log_paths,
                role_ids,
                log_level="WARN",
                search_timeout=TEN_SECONDS_IN_MILLIS,
                query=None,
                role_result_limit=None,
                total_byte_limit=None,
                host=None,
                port=None):
  """
  Performs a log search and returns a a dictionary containing search metadata
  and all log events found.

         start_time: Required. The earliest time for which to return log events, in seconds since the epoch.
           end_time: Required. The latest time for which to return log events, in seconds since the epoch
          log_types: Required. A list of log types. Valid log types include: LOG4J, HUE, GLOG.
          log_paths: Required. A list of log paths. e.g. ["/var/log/hadoop/hdfs.log", "/var/log/hue/runcpserver.log"]
           role_ids: Required. A list of role ids. e.g. [23, 45, 1]
          log_level: Defaults to WARN. A log4j log level.
     search_timeout: Defaults to 10 seconds. Time to stream logs before raising a timeout, in milliseconds.
              query: Optional. If set to a string, all results returned with contain query.
  role_result_limit: Optional. A cap on the number of log events returned per role.
   total_byte_limit: Optional. A cap on log event bytes returned.
               host: Optional. A host to include in search metadata (e.g. this host)
               port: Optional. A port to include in search metadata (e.g. the port this data
                     will be ultimately served from)
  """
  results = []
  def append_event(event, first_event):
    results.append(event)
    return get_event_dict_size(event)

  search_data = stream_logs(event_callback=append_event,
                            start_time=start_time,
                            end_time=end_time,
                            log_types=log_types,
                            log_paths=log_paths,
                            role_ids=role_ids,
                            log_level=log_level,
                            search_timeout=search_timeout,
                            query=query,
                            role_result_limit=role_result_limit,
                            total_byte_limit=total_byte_limit,
                            host=host,
                            port=port)

  search_data["events"] = results

  return {
    "content_type": "application/json",
    "data": json.dumps(search_data, default=datetime_converter)
  }

def search_logs_compressed(start_time,
                           end_time,
                           log_types,
                           log_paths,
                           role_ids,
                           log_level="WARN",
                           search_timeout=TEN_SECONDS_IN_MILLIS,
                           query=None,
                           role_result_limit=None,
                           total_byte_limit=None,
                           host=None,
                           port=None):
  """
  See search_logs(...) for an explanation of these parameters.

  This will return a JSON results dictionary mapping:

  - "content_type": the content type of the compressed logs
  - "path": the path to the compressed logs

  The file at "path" will be a zip file containing:

  - log_events.json.gz: contains gzip-compressed json log events
  - meta.json: contains search metadata

  Why the gzip/zip awkwardness? Python's zipfile library doesn't support writing
  to the same internal zip entry over multiple calls. If compression is enabled,
  presumably the number of raw events will be rather large, so we don't want to
  bring them all into memory at once to write one entry. Instead, we want to
  compress events as we get them and write them to a buffer. The only way
  this is possible with a ZipFile object is by creating a new file entry for
  each event, which isn't great.

  So, we compress the log events using gzip and then archive the compressed
  events and search metadata with zip, rather than use the zipfile library for
  both archiving and compression.
  """

  events_path = None
  events_fd = None
  events_gzip_fd = None
  archive_path = None
  archive_fd = None
  archive_zip_fd = None

  try:
    (events_fd, events_path) = tempfile.mkstemp()
    events_fd = os.fdopen(events_fd, "w")
    events_gzip_fd = gzip.GzipFile(mode="w", compresslevel=9, fileobj=events_fd)
    events_gzip_fd.write("[") # Open square bracket to open a JSON array.

    def write_event(event, first_event):
      before_length = events_fd.tell()
      if not first_event:
        events_gzip_fd.write(",") # Comma to delimit JSON event dicts.
      events_gzip_fd.write(json.dumps(event, default=datetime_converter, indent=True))
      events_gzip_fd.write("\n") # Newline between events.
      events_gzip_fd.flush()
      return events_fd.tell() - before_length

    search_data = stream_logs(
      event_callback=write_event,
      start_time=start_time,
      end_time=end_time,
      log_types=log_types,
      log_paths=log_paths,
      role_ids=role_ids,
      log_level=log_level,
      search_timeout=search_timeout,
      query=query,
      role_result_limit=role_result_limit,
      total_byte_limit=total_byte_limit,
      host=host,
      port=port)

    events_gzip_fd.write("]") # Close square bracket to close a JSON array.
    events_gzip_fd.close()
    events_fd.close()

    # Archive search metadata and results.
    (archive_fd, archive_path) = tempfile.mkstemp()
    archive_fd = os.fdopen(archive_fd, "w")
    archive_zip_fd = zipfile.ZipFile(archive_fd, "w")
    archive_zip_fd.writestr("meta.json",
                            json.dumps(search_data, default=datetime_converter))
    archive_zip_fd.write(events_path, "log_events.json.gz")
    archive_zip_fd.close()
    archive_fd.close()

    return {
      "content_type": "application/zip",
      "path": archive_path
    }
  finally:
    util.close_quietly(events_gzip_fd)
    util.close_quietly(events_fd)
    util.close_quietly(archive_zip_fd)
    util.close_quietly(archive_fd)
    util.delete_quietly(events_path)
    # Don't delete archive_path, since it's the "return value"

def stream_logs(event_callback,
                start_time,
                end_time,
                log_types,
                log_paths,
                role_ids,
                log_level="WARN",
                search_timeout=TEN_SECONDS_IN_MILLIS,
                query=None,
                role_result_limit=None,
                total_byte_limit=None,
                host=None,
                port=None):
  """
  See search_logs(...) for an explanation of these parameters.

  The only extra parameter here is event_callback, which should be a function
  that takes one parameter, an event dictionary. This callback is invoked
  on every event that the log streamers return, and the callback should return
  the number of bytes written.

  This function returns a dictionary containing search metadata.
  """

  assert(None not in (event_callback, start_time, end_time, log_paths, log_types, role_ids))
  assert(len(log_paths) == len(log_types) == len(role_ids))

  # measure cost in time, cpu, and io
  start_timer = time.time()
  start_user, start_system = proc.get_cpu_times()
  read_count, write_count, read_bytes, write_bytes = get_io_counters()

  # parse out all the params.
  start_datetime = datetime_from_millis(start_time)
  end_datetime = datetime_from_millis(end_time)
  log_level = parse_log_level(log_level)
  timeout = time.time() + int(search_timeout)/1000.0

  messages = []
  try:
    log_streamer_role_id_tuples = []

    for (log_path, log_type, role_id) in zip(log_paths, log_types, role_ids):
      # TODO: security issue: request can cause read of any dir
      log_dir, log_name = os.path.split(log_path)

      try:
        validate_log_dir(log_dir)
        validate_log_type(log_type)
      except Exception, e:
        messages.append(e.args[0])
        continue


      log_streamer = get_log_streamer(
        log_type, log_dir, log_name, log_level, start_datetime, end_datetime, query, timeout)

      log_streamer_role_id_tuples.append((log_streamer, role_id))

    byte_count = 0
    event_count = 0
    first_event = True

    if total_byte_limit:
      # To honor a total byte limit while getting as many logs as possible for
      # each role, we cycle through each streamer and collect one event at a time
      # from each.
      stream_role_id_tuples = [(log_streamer.stream(), role_id) for
                               log_streamer, role_id in
                               log_streamer_role_id_tuples]
      while True:
        if len(stream_role_id_tuples) == 0 or \
           byte_count >= total_byte_limit or \
           (role_result_limit and event_count >= role_result_limit):
          break

        to_remove = []

        for i, (stream, role_id)  in enumerate(stream_role_id_tuples):
          try:
            event = stream.next()
          except StopIteration:
            to_remove.append((stream, role_id))
            continue

          event_dict = event2Dict(event, role_id)
          byte_count += event_callback(event_dict, first_event)
          if byte_count >= total_byte_limit:
            break
          first_event = False

        event_count += 1

        for item in to_remove:
          stream_role_id_tuples.remove(item)
    else:
      # Without a byte limit, we collect events from each streamer until the
      # role result limit is met or the streamer runs out of events.
      #
      # The reason we implement this strategy in addition to the total-byte-limit-
      # honoring strategy is for "backwards-compatibility", so that the ordering
      # of results shown on the front-end doesn't change.
      for log_streamer, role_id in log_streamer_role_id_tuples:
        event_count = 0
        for event in log_streamer.stream():
          event_dict = event2Dict(event, role_id)
          event_callback(event_dict, first_event)
          event_count += 1
          if role_result_limit and event_count >= role_result_limit:
            break
          first_event = False

  except TimeoutException:
    messages.append(TIMEOUT_MESSAGE)

  end_user, end_system = proc.get_cpu_times()
  end = get_io_counters()

  search_data = dict(
    success=True,
    messages=messages,
    cost=dict(
      time=time.time() - start_timer,
      io=dict(
        read=dict(
          count=end.read_count - read_count,
          bytes=end.read_bytes - read_bytes),
        write=dict(
          count=end.write_count - write_count,
          bytes=end.write_bytes - write_bytes)),
      cpu=dict(
        user=end_user - start_user,
        system=end_system - start_system)),
    host=host,
    port=port,
    start=start_datetime,
    end=end_datetime)

  return search_data



def log_with_context(path, offset, direction, num_events, log_type, host, port):
  """
  Returns JSON for log events in the lines starting from a given offset or the end
  of the file if no offset is provided.

  Arguments:

        path: Required. Log file path.
      offset: Optional. Byte offset into the log file, set to end of file if not provided.
   direciton: Optional. Can either be "previous" or "next" to indicate where to go from the offset.
  if it's not provided logs surrounding the offset will be retrieved.
  num_events: Required. Number of log messages to get.
    log_type: Required. Can be either LOG4J, GLOG or HUE.
        host: Required. Agent host to be attached to the response.
        port: Required. Agent port to be attached to the response.
  """
  f = LogFile(path)
  end_offset = f.get_end_offset()

  if offset == None:
    offset = end_offset

  if direction == NEXT:
    max_prev_events = 0
  elif direction == PREVIOUS:
    max_prev_events = num_events
  else:
    max_prev_events = num_events / 2

  if log_type == LOG4J:
    event_parser = Log4JEventParser()
  elif log_type == HUE:
    event_parser = HueEventParser()
  elif log_type == GLOG:
    event_parser = GLogEventParser()


  events = []
  handle = f.open_log_file()
  try:

    if direction is None or direction == PREVIOUS:
      reader = LogFileEventReader(
        event_parser=event_parser,
        file_handle=handle,
        file_path=path,
        start_offset=offset)

      while True:
        # go backward
        event = reader.get_prev_event()
        if event is not None and event.get_offset() >= 0 and len(events) < max_prev_events:
          events.append(event2Dict(event))
        else:
          break

    if direction is None or direction == NEXT:
      reader = LogFileEventReader(
        event_parser=event_parser,
        file_handle=handle,
        file_path=path,
        start_offset=offset)
      while True:
        # go forward
        event = reader.get_next_event()
        if event is not None and event.get_offset() < end_offset and len(events) < num_events:
          events.append(event2Dict(event))
        else:
          break

  except:
      # python 2.4 doesn't allow yield in a try finally block, so instead
      # we catch all exceptions, close the handle, and re-raise the exception.
      handle.close()
      raise
  handle.close()

  return dict(
    content_type="application/json", # so cherrypy knows what to use

    # json that is returned
    data=json.dumps(dict(
      events=events,
      success=True,
      host=host,
      port=port),
    default=datetime_converter))


def event2Dict(event, role_id=None):
  """
  Turns LogEvent into a dictionary representation.
  """

  return dict(
    offset=event.get_offset(),
    path=event.get_file_path(),
    time=event.get_datetime(),
    loglevel=event.get_log_level(),
    message=event.get_message(),
    source=event.get_source(),
    roleId=role_id)

def get_event_dict_size(event):
  """Instead of getting the length of the JSON serialization, sum all the
  event dict field and add 8 bytes (a comma between each field + two curly
  brackets)."""
  return len(str(event["offset"])) + \
      len(str(event["path"])) + \
      len(str(datetime_converter(event["time"]))) + \
      len(str(event["loglevel"])) + \
      len(str(event["message"])) + \
      len(str(event["source"])) + \
      len(str(event["roleId"])) + \
      + 8

def get_io_counters():
  try:
    return proc.get_io_counters()
  except psutil.NoSuchProcess:
    # We don't support I/O metrics on linux machines with kernels that don't
    # support I/O Stats (that don't have /proc/<pid>/io)
    return ntuple_io(0,0,0,0)


def get_query_regex(query):
  """return a compiled regex for the query. If the query is escaped using java
  regex pattern escaping '\Q<query>\E' then we escape the regex, otherwise
  compile the query as is"""
  m = quote_regex.match(query)
  if m is None:
    # this query is not quoted, compile as is
    return  re.compile(query, re.MULTILINE | re.IGNORECASE)
  else:
    # this is a quoted query that should be escaped
    return re.compile(re.escape(m.group(1)), re.MULTILINE | re.IGNORECASE)

