#!/usr/bin/env python
# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import logging
import os
import re

import cmf.version

SPECIAL_SCRIPTS = {'mr1' : '/bin/hadoop',
                   'mr2' : '/bin/mapred',
                   'httpfs' : '/sbin/httpfs.sh',
                   'zookeeper' : '/bin/zkServer.sh',
                   'hue' : '/build/env/bin/hue',
                   'flume' : '/bin/flume-ng',
                   'impala' : '/sbin/impalad'}

AVAILABLE_CDH_VERSIONS = ['cdh3', 'cdh4']

LOG = logging.getLogger(__name__)

VERSION_FILE_CDH_KEY='cloudera.cdh.release'
VERSION_FILE_COMPONENT_KEY='cloudera.pkg.version'

CDH_VERSION_KEY='cdh_version'
COMPONENT_VERSION_KEY='component_version'

class VersionDetector:

  @classmethod
  def _get_key_values_from_file(cls, version_file):
    """
    Returns a dict where lines in version_file are split as key=value.
    """
    version_dict = dict()
    f = open(version_file)
    contents = f.read().strip().splitlines()
    f.close()
    for line in contents:
      if line.find('=') >= 0:
        kv_array = line.split('=')
        if len(kv_array) == 2:
          version_dict[kv_array[0]] = kv_array[1]
    return version_dict

  @classmethod
  def _get_version_for_cdh_component(cls, cdh_component_home, is_cm=False):
    """
    Returns version for the CDH component.
    """
    version_file = cdh_component_home + "/cloudera/cdh_version.properties"
    if is_cm:
      version_file = cdh_component_home + "/cloudera/cm_version.properties"
    version_dict = dict()

    if os.path.exists(version_file):
      file_dict = cls._get_key_values_from_file(version_file)
      if VERSION_FILE_CDH_KEY in file_dict and \
           VERSION_FILE_COMPONENT_KEY in file_dict:
        try:
          # The first 4 chars represent CDH version.
          # This is good only till we get to CDH10.
          cdh_version = file_dict[VERSION_FILE_CDH_KEY][:4]
          if cdh_version in AVAILABLE_CDH_VERSIONS:
            # Return the correct version
            version_dict[CDH_VERSION_KEY] = cdh_version.upper()
            version_dict[COMPONENT_VERSION_KEY] = file_dict[VERSION_FILE_COMPONENT_KEY]
            return version_dict
          raise Exception("Unsupported version: " + cdh_version)
        except Exception, e:
          LOG.exception("Error while reading version file: " + version_file \
                        + " value: " + file_dict[VERSION_FILE_CDH_KEY] \
                        + "\nexception: " + str(e))
      # Version file is present but isn't in correct format or has unsupported version
      version_dict[CDH_VERSION_KEY] = version_dict[COMPONENT_VERSION_KEY] = 'BAD_VERSION_FILE'
    else:
      # Version file is not present
      version_dict[CDH_VERSION_KEY] = version_dict[COMPONENT_VERSION_KEY] = 'NO_VERSION_FILE'

    return version_dict

  @classmethod
  def _check_file_exists_with_pattern(cls, directory, filename_pattern):
    """
    Returns true if a file with the given filename pattern exists in the
    specified directory.
    """
    for f in os.listdir(directory):
      if re.match(filename_pattern, f) is not None:
        return True
    return False

  @classmethod
  def get_component_versions(cls, supervisee_environment):
    """
    The main loop which goes through all the environment variables of
    type CDH_xxx_HOME and finds out the versions for the corresponding components.
    """
    installed_versions = []

    # Combine the OS environment and the supervisee environment
    all_env = os.environ.copy()
    all_env.update(supervisee_environment)
    for k, v in all_env.iteritems():
      cdh_match = re.match("CDH_(.+)_HOME", k)
      # ignore any env variable not in CDH_XXX_HOME format
      if cdh_match is None:
        continue

      cdh_component = cdh_match.group(1).lower()

      cdh_component_script = v + "/bin/" + cdh_component
      if cdh_component in SPECIAL_SCRIPTS:
        cdh_component_script = v + SPECIAL_SCRIPTS[cdh_component]

      version_info = dict()
      # returned if even the script doesn't exist for the component
      version_info[CDH_VERSION_KEY] = version_info[COMPONENT_VERSION_KEY] = 'NO_SCRIPT'

      try:
        if cdh_component == 'hue_plugins':
          # Even though hue-plugins are a CDH component, they share their home with
          # hadoop and we already had a cdh-version.properties for that. So, we have
          # to use cm-version.properties for hue-plugins.
          if cls._check_file_exists_with_pattern(os.path.join(v, "lib"), "hue-plugins(.+).jar"):
            version_info = cls._get_version_for_cdh_component(v, True)
        elif os.path.exists(cdh_component_script):
          version_info = cls._get_version_for_cdh_component(v)
      except Exception, e:
        LOG.exception("Error while reading version file.")

      version_info['name'] = cdh_component
      installed_versions.append(version_info)

    # Add in mgmt daemons version
    if "CMF_AGENT_MGMT_HOME" in all_env:
      version_info = dict()
      version_info["name"] = "cm_daemons"
      version_info[CDH_VERSION_KEY] = "NOT_APPLICABLE"
      version_info[COMPONENT_VERSION_KEY] = "NO_SCRIPT"
      try:
        home = all_env.get("CMF_AGENT_MGMT_HOME")
        file_dict = cls._get_key_values_from_file(os.path.join(home, "cloudera/cm_version.properties"))
        version_info[COMPONENT_VERSION_KEY] = file_dict["version"]
      except Exception, e:
        LOG.exception("Error while reading version file.")
      installed_versions.append(version_info)

    # Add in the agent version
    version_info = dict()
    version_info["name"] = "cm_agent"
    version_info[CDH_VERSION_KEY] = "NOT_APPLICABLE"
    version_info[COMPONENT_VERSION_KEY] = cmf.version.VERSION
    installed_versions.append(version_info)
    return installed_versions
