#!/bin/bash
#
# (c) Copyright 2011-2012 Cloudera, Inc.
#
# This script initializes SCM and Monitoring services "embedded" database.
# This involves:
#   * Running PostgreSQL's "initdb" to create an empty database server.
#   * Updating database properties to accept connection remotely.
#   * Updating max number of connections that server can open.
#   * Starting the server.
#   * Creating empty databases for SCM, SMON, AMON, HMON and RMAN.
#   * Stopping the server.

# Display an error and either fail or continue executing.
fail_or_continue()
{
    local RET=$1
    local STR=$2

    if [[ $RET -ne 0 ]]; then
        if [[ -z $STR ]]; then
            STR="--> Error $RET"
        fi
        echo "$STR, giving up"
        exit $RET
    fi
}

create_databases()
{
  export PGPASSWORD=$PASSWORD
  local DB_CMD="psql --host=localhost --port=$DB_PORT --username=$USER --dbname=postgres -v ON_ERROR_STOP=1"

  # Create database for all roles
  while read line
  do
    set $line
    $DB_CMD --command "CREATE ROLE $1 LOGIN PASSWORD '$2';"
    fail_or_continue $? "Unable to create database role $1"
    $DB_CMD --command "CREATE DATABASE $1 OWNER $1;"
    fail_or_continue $? "Unable to create database $1"
  done <<EOF
scm $SCM_PASSWORD
smon $SMON_PASSWORD
amon $AMON_PASSWORD
hmon $HMON_PASSWORD
rman $RMAN_PASSWORD
EOF

  fail_or_continue $? "Could not create database $SCM_DATABASE"
}

write_config_file()
{
  # Write out the configuration file.
  echo "Creating SCM configuration file: $DB_PROP_FILE"

  cat > $DB_PROP_FILE << EOF
# Auto-generated by `basename $0`
#
# `date`
#
# These are database settings for CM Manager
#
com.cloudera.cmf.db.type=postgresql
com.cloudera.cmf.db.host=localhost:$DB_PORT
com.cloudera.cmf.db.name=scm
com.cloudera.cmf.db.user=scm
com.cloudera.cmf.db.password=$SCM_PASSWORD
EOF

  fail_or_continue $? "Error creating file $DB_PROP_FILE"
  echo "Created db properties file $DB_PROP_FILE"

  echo "Creating Mgmt configuration file: $MGMT_DB_PROP_FILE"

  cat > $MGMT_DB_PROP_FILE << EOF
# Auto-generated by `basename $0`
#
# `date`
#
# These are database settings for the
# Cloudera Manager Management Services (Full Edition only),
# and are used during the creation wizard. The source of
# record for these settings is the database.
#
EOF

  while read line
  do
    set $line
    # $1 is MgmtServiceHandler.RoleNames Enum value
    local PREFIX="com.cloudera.cmf.$1.db"

    cat >> $MGMT_DB_PROP_FILE <<EOF
$PREFIX.type=postgresql
$PREFIX.host=$DB_HOSTPORT
$PREFIX.name=$2
$PREFIX.user=$2
$PREFIX.password=$3
EOF
  fail_or_continue $? "Error creating file $MGMT_DB_PROP_FILE"

  done <<EOF
SERVICEMONITOR smon $SMON_PASSWORD
ACTIVITYMONITOR amon $AMON_PASSWORD
HOSTMONITOR hmon $HMON_PASSWORD
REPORTSMANAGER rman $RMAN_PASSWORD
EOF

  echo "Created db properties file $MGMT_DB_PROP_FILE"
}

cd
DB_HOST=$(hostname -f)
DB_PORT=${DB_PORT:-7432}
DB_HOSTPORT="$DB_HOST:$DB_PORT"
DB_PROP_FILE=/etc/cloudera-scm-server/db.properties
MGMT_DB_PROP_FILE=/etc/cloudera-scm-server/db.mgmt.properties
NOW=$(date)

if [[ $# -ne 2 ]]; then
    echo "Usage: $0 <data_dir> <server_logfile>"
    exit 1
fi

DATA_DIR=$1
SERVER_LOGFILE=$2


if [[ -n `find "$DATA_DIR" -maxdepth 0 -not -empty` ]]; then
    echo "$DATA_DIR is non-empty; perhaps the database was already initialized?"
    exit 1
fi

# Create the directory if it doesn't already exist
install -m 700 -d $DATA_DIR
fail_or_continue $? "Could not create $DATA_DIR"

# Create an alphanumeric 10 character password
PASSWORD=$(perl -le 'print map { ("a".."z", "A".."Z", 0..9)[rand 62] } 1..10')
SCM_PASSWORD=$(perl -le 'print map { ("a".."z", "A".."Z", 0..9)[rand 62] } 1..10')
SMON_PASSWORD=$(perl -le 'print map { ("a".."z", "A".."Z", 0..9)[rand 62] } 1..10')
AMON_PASSWORD=$(perl -le 'print map { ("a".."z", "A".."Z", 0..9)[rand 62] } 1..10')
HMON_PASSWORD=$(perl -le 'print map { ("a".."z", "A".."Z", 0..9)[rand 62] } 1..10')
RMAN_PASSWORD=$(perl -le 'print map { ("a".."z", "A".."Z", 0..9)[rand 62] } 1..10')

PASSWORD_TMP_FILE=$(mktemp /tmp/XXXXXXXX)
cat > $PASSWORD_TMP_FILE <<EOF
$PASSWORD

The password above was generated by $0 (part of the cloudera-scm-server-db package)
and is the password for the user 'cloudera-scm' for the database in the current directory.

Generated at $NOW.
EOF

initdb --pgdata "$DATA_DIR" --encoding=UTF8 --locale=en_US.UTF8 --auth=md5 --username="$USER" --pwfile="$PASSWORD_TMP_FILE"
fail_or_continue $? "Could not initialize database server.
This usually means that your PostgreSQL installation failed or isn't working properly.
PostgreSQL is installed using the set of repositories found on this machine. Please
ensure that PostgreSQL can be installed. Please also uninstall any other instances of
PostgreSQL and then try again."

# We save the password so that the database is re-usable later.
# Note that we couldn't create this file initially, because initdb
# requires an empty directory to work in.
PASSWORD_FILE="$DATA_DIR/generated_password.txt"
install -m 0600 $PASSWORD_TMP_FILE $PASSWORD_FILE

# Re-configure the listen address and port, since the postgresql-server
# package may be using the default postgres port and listen address.
# Though typically the default configs don't specify a
# port, we try to remove it anyway.
sed -e '/^port\s*=/d' -i $DATA_DIR/postgresql.conf

# Listen on all IP addresses, as monitoring services may reside on
# different machines on the LAN.
sed -e '/^listen_addresses\s*=/d' -i $DATA_DIR/postgresql.conf

# Bump up max connections to server and shared buffer space that connections
# need. shared_buffers should be at least 2 * max_connections.
sed -e '/^max_connections\s*=/d' -i $DATA_DIR/postgresql.conf
sed -e '/^shared_buffers\s*=/d' -i $DATA_DIR/postgresql.conf


# Prepend to the file
TMPFILE=$(mktemp /tmp/XXXXXXXX)
cat > $TMPFILE << EOF
#########################################
# === Generated by cloudera-scm-server-db at $NOW
port = $DB_PORT
listen_addresses = '*'
max_connections = 600
shared_buffers = 1500
#########################################
EOF
cat $DATA_DIR/postgresql.conf >> $TMPFILE
cat $TMPFILE > $DATA_DIR/postgresql.conf

# Add a entry to pg_hba.conf file to accept connections from any host
echo "# Accept connection from all hosts" >> $DATA_DIR/pg_hba.conf
echo "host     smon     smon     0.0.0.0/0     md5" >> $DATA_DIR/pg_hba.conf
echo "host     amon     amon     0.0.0.0/0     md5" >> $DATA_DIR/pg_hba.conf
echo "host     hmon     hmon     0.0.0.0/0     md5" >> $DATA_DIR/pg_hba.conf
echo "host     rman     rman     0.0.0.0/0     md5" >> $DATA_DIR/pg_hba.conf

# Start the database server (-w waits for startup)
pg_ctl start -w -l "$SERVER_LOGFILE" -D "$DATA_DIR" -o "$EXTRA_PG_ARGS"
fail_or_continue $? "Could not start database server"

# Backup db.properties
if [[ -e $DB_PROP_FILE ]]; then
    cp $DB_PROP_FILE $DB_PROP_FILE.$(date +%Y%m%d-%H%M%S)
fi

# Create the databases
create_databases
write_config_file

# Stop the server
pg_ctl stop -w -l "$SERVER_LOGFILE" -D "$DATA_DIR" -o "$EXTRA_PG_ARGS"
fail_or_continue $? "Could not stop database server"

echo "Done."
