// (c) Copyright 2011-2012 Cloudera, Inc. All rights reserved.
define([
	"cloudera/chart/ContextChartSet2",
	"cloudera/chart/ChartUtil"
], function(ContextChartSet2, ChartUtil) {

return ContextChartSet2.extend({

	/**
	 * Adding chart descriptors and contexts.
	 *
	 * contexts - contexts describes what we need to plot time series for.
	 * [{
	 *   id: ...,
	 *   type: ...
	 * }]
	 * chartDescriptors - chart descriptors describes how to display a chart,
	 * and how to display a metric in a chart.
	 * [{
	 *   metric: {
	 *     id: 3, // metric id
	 *     context: // ContextType
	 *     desc: ...,
	 *   },
	 *   chart: {
	 *     id: ..., // chart id
	 *     title: ..., // chart title
	 *     ...
	 *   }
	 * }]
	 * skipRender - (optional, default false). If true, this.render() will not
	 * be called, which can be expensive. This allows us to update the chart
	 * settings with multiple API calls, and update AnyChart's settings at once.
	 */
	initializeData: function() {
		this._super.apply(this, arguments);
		this.contexts = [];
		this.chartDescriptors = [];
		this.serviceName = this.options.serviceName;

		try {
			var skipRender = true;
			this.setChartDescriptors(this.options.chartDescriptors, skipRender);
			this.setContexts(this.options.contexts);
		} catch (ex) {
			console.log(ex);
		}
	},

	/**
	 * Generates the parameters for the URL parameters of getting the Chart's timeseries.
	 * FIXME: This generates really long URLs and need to be fixed.
	 * FIXME: This also doesn't do URL escaping properly. so if any metricId, contextId
	 * contain special characters,
	 */
	getMetricParams: function(index, metric, contextType, contextId) {
		var proxyCM = "proxyContextMetrics[" + index + "]";

		var buf = [];
		buf.push(proxyCM + ".metric.metricId=" + metric.id);
		buf.push(proxyCM + ".metric.derivativeType=" + metric.derivative);
		buf.push(proxyCM + ".metric.derivativeScale=" + metric.scale);
		buf.push(proxyCM + ".metric.units=" + metric.units);
		buf.push(proxyCM + ".context.contextType=" + contextType.toUpperCase());
		buf.push(proxyCM + ".context.contextId=" + contextId);

		return buf.join("&");
	},

	/**
	 * Generates the URL parameters for all the contexts, metrics combination.
	 */
	_generateChartsPostData: function(startDate, endDate) {
		var buf = [];
		buf.push("startTime=" + startDate.getTime());
		buf.push("endTime=" + endDate.getTime());

		var i, j, k = 0;
		for (i = 0; i < this.contexts.length; ++i) {
			var context = this.contexts[i];
			for (j = 0; j < this.chartDescriptors.length; ++j) {
				var chartDescriptor = this.chartDescriptors[j];
				if (chartDescriptor.metric.context === context.type) {
					var params = this.getMetricParams(k, chartDescriptor.metric, context.type, context.id);
					buf.push(params);
					k++;
				}
			}
		}
		return buf.join("&");
	},

	/**
	 * Generates the URL parameters for the event flags.
	 */
	_generateEventsPostData: function(startDate, endDate) {
		var buf = [];
		if (this.serviceName) {
			buf.push("service=" + this.serviceName);
		}
		buf.push("serviceType=" + "MAPREDUCE");
		buf.push("timeUpper=" + endDate.getTime());
		buf.push("timeLower=" + startDate.getTime());
		buf.push("summarise=on");

		var duration = this.delta(endDate, startDate);
		// by default, we want to see at most 10 events per category for each fetch.
		var summariseSize = duration / 10;
		if (summariseSize > 0) {
			buf.push("summariseSize=" + summariseSize);
		}
		buf.push("summariseGroupBy=CATEGORY");
		buf.push("summariseBy=SEVERITY");

		return buf.join("&");
	},

	/**
	 * Adds a single context.
	 */
	addContext: function(context) {
		this.addUniqueEntry(this.contexts, context);

		var i;
		for (i = 0; i < this.chartDescriptors.length; i += 1) {
			var chartDescriptor = this.chartDescriptors[i];
			this._addRemoveTimeSeries(context, chartDescriptor, true);
		}
	},

	/**
	 * Removes a single context.
	 */
	removeContext: function(context) {
		var i;
		for (i = 0; i < this.chartDescriptors.length; i += 1) {
			var chartDescriptor = this.chartDescriptors[i];
			this._addRemoveTimeSeries(context, chartDescriptor, false);
		}
		this.removeEntryById(this.contexts, context.id);
	},

	/**
	 * @return a context if it is in the system, null if not found.
	 */
	getContext: function(context) {
		return this.getEntry(this.contexts, context.id);
	},

	/**
	 * Adds a list of contexts.
	 */
	addContexts: function(contexts, skipRender) {
		this._addRemoveContexts(contexts, true, skipRender);
	},

	/**
	 * Removes a list of contexts.
	 */
	removeContexts: function(contexts, skipRender) {
		this._addRemoveContexts(contexts, false, skipRender);
	},

	/**
	 * Utility method to add/remove contexts.
	 */
	_addRemoveContexts: function(contexts, isAdd, skipRender) {
		var i;
		for (i = 0; i < contexts.length; i += 1) {
			var context = contexts[i];
			if (isAdd) {
				this.addContext(context);
			} else {
				this.removeContext(context);
			}
		}
		if (!skipRender) {
			this.render();
		}
	},

	/**
	 * Replaces the existing context list with a new list.
	 */
	setContexts: function(contexts, skipRender) {
		var comparator = function(c1, c2) {
			return (c1.id === c2.id && c1.type === c2.type);
		};
		var addRemoves = ChartUtil.getAddRemoves(this.contexts, contexts, comparator);
		var localSkipRender = true;
		this.removeContexts(addRemoves.removes, localSkipRender);
		this.addContexts(addRemoves.adds, localSkipRender);
		if (!skipRender) {
			this.render();
		}
	},

	/**
	 * Adds a chart descriptor.
	 */
	addChartDescriptor: function(chartDescriptor) {
		var chartId = this.getChartId(chartDescriptor.chart);
		var chartTitle = chartDescriptor.chart.title;
		// Assume the units of all the time series on a single chart
		// are identical. 
		var units = chartDescriptor.metric.units;
		var charts = this.getCharts();

		if (!this.getChartDescriptor(chartDescriptor)) {
			this.addEntry(this.chartDescriptors, chartDescriptor);
		}
		// Note: this.createChart is quick.
		// So even if the chart already exist,
		// it is fine.
		this.addUniqueEntry(charts, this.createChart(chartId, chartTitle, units));
		this._settingsChanged = true;

		var i;
		for (i = 0; i < this.contexts.length; i += 1) {
			var context = this.contexts[i];
			this._addRemoveTimeSeries(context, chartDescriptor, true);
		}
	},

	/**
	 * Removes a chart descriptor.
	 */
	removeChartDescriptor: function(chartDescriptor) {
		var i;
		for (i = 0; i < this.contexts.length; i += 1) {
			var context = this.contexts[i];
			this._addRemoveTimeSeries(context, chartDescriptor, false);
		}

		var comparator = this._getChartDescriptorComparator(chartDescriptor);
		this.removeEntryByPredicate(this.chartDescriptors, comparator);

		// Remove the chart if there are no more metrics mapped to it.
		var oldChartId = chartDescriptor.chart.id;
		var canRemoveChart = true;

		for (i = 0; i < this.chartDescriptors.length; i += 1) {
			var d = this.chartDescriptors[i];
			if (d.chart.id === oldChartId) {
				// found a match, don't remove the chartId.
				canRemoveChart = false;
			}
		}

		if (canRemoveChart) {
			var chartId = this.getChartId(chartDescriptor.chart);
			var charts = this.getCharts();
			this.removeEntryById(charts, chartId);
			this._settingsChanged = true;
		}
	},

	/**
	 * Utility to add/remove a context/chart descriptor combination.
	 * This defines a single time series.
	 *
	 * @param context - context.
	 * @param chartDescriptor - chart descriptor.
	 * @param skipRender - (optional, default false). If true,
	 * this.render() will not be called, which can be
	 * expensive.
	 */
	_addRemoveTimeSeries: function(context, chartDescriptor, isAdd) {
		var timeSeriesDescriptor = {
			context: context,
			metric: chartDescriptor.metric,
			chart: chartDescriptor.chart
		};
		if (isAdd) {
			this.addTimeSeries(timeSeriesDescriptor);
		} else {
			this.removeTimeSeries(timeSeriesDescriptor);
		}
	},

	_getChartDescriptorComparator: function(chartDescriptor) {
		return function(element) {
			return (element.metric.id === chartDescriptor.metric.id &&
						  element.metric.context === chartDescriptor.metric.context);
		};
	},

	/**
	 * @return a chartDescriptor if it is present in the system.
	 */
	getChartDescriptor: function(chartDescriptor) {
		var comparator = this._getChartDescriptorComparator(chartDescriptor);
		return this.getEntryByPredicate(this.metrics, comparator);
	},

	/**
	 * Adds a list of chart descriptors.
	 */
	addChartDescriptors: function(chartDescriptors, skipRender) {
		this._addRemoveChartDescriptors(chartDescriptors, true, skipRender);
	},

	/**
	 * Removes a list of chart descriptors.
	 */
	removeChartDescriptors: function(chartDescriptors, skipRender) {
		this._addRemoveChartDescriptors(chartDescriptors, false, skipRender);
	},

	/**
	 * Utility to add/remove a chart descriptors.
	 *
	 * @param chartDescriptors - a list of chart descriptors to add/remove.
	 * @param isAdd - true to add, false to remove.
	 * @param skipRender - (optional, default false). If true,
	 * this.render() will not be called, which can be
	 * expensive.
	 */
	_addRemoveChartDescriptors: function(chartDescriptors, isAdd, skipRender) {
		var i = 0;
		for (i = 0; i < chartDescriptors.length; i += 1) {
			var chartDescriptor = chartDescriptors[i];
			if (isAdd) {
				this.addChartDescriptor(chartDescriptor);
			} else {
				this.removeChartDescriptor(chartDescriptor);
			}
		}
		if (!skipRender) {
			this.render();
		}
	},

	/**
	 * Replaces the existing chart descriptors list with a new list.
	 */
	setChartDescriptors: function(chartDescriptors, skipRender) {
		var comparator = function(m1, m2) {
			return (m1.metric.id === m2.metric.id && m1.metric.context === m2.metric.context);
		};
		var addRemoves = ChartUtil.getAddRemoves(this.chartDescriptors, chartDescriptors, comparator);
		var localSkipRender = true;
		this.removeChartDescriptors(addRemoves.removes, localSkipRender);
		this.addChartDescriptors(addRemoves.adds, localSkipRender);
		if (!skipRender) {
			this.render();
		}
	}
});
});
